import logging
import logging.config
import logging.handlers
from sys import stdout
from cached_property import cached_property
from egcg_core.config import cfg


class LoggingConfiguration:
    """Stores Loggers, Formatters and Handlers"""

    default_fmt = '[%(asctime)s][%(name)s][%(levelname)s] %(message)s'
    default_datefmt = '%Y-%b-%d %H:%M:%S'

    def __init__(self, config):
        self.cfg = config
        self.blank_formatter = logging.Formatter()
        self.handlers = set()
        self.loggers = {}
        self._log_level = logging.INFO

    @cached_property
    def formatter(self):
        return self.default_formatter

    @cached_property
    def default_formatter(self):
        return logging.Formatter(
            fmt=self.cfg.get('format', self.default_fmt),
            datefmt=self.cfg.get('datefmt', self.default_datefmt)
        )

    def get_logger(self, name, level=logging.NOTSET):
        """
        Return a logging.Logger object with formatters and handlers added.
        :param name: Name to assign to the logger (usually __name__)
        :param int level: Log level to assign to the logger upon creation
        """
        if name in self.loggers:
            logger = self.loggers[name]
        else:
            logger = logging.getLogger(name)
            self.loggers[name] = logger

        logger.setLevel(level or self._log_level)
        for h in self.handlers:
            logger.addHandler(h)

        return logger

    def add_handler(self, handler, level=logging.NOTSET):
        """
        Add a created handler, set its format/level if needed and register all loggers to it
        :param logging.Handler handler:
        :param int level: Log level to assign to the created handler
        """
        handler.setLevel(level or self._log_level)
        handler.setFormatter(self.formatter)
        for name in self.loggers:
            self.loggers[name].addHandler(handler)
        self.handlers.add(handler)

    def add_stdout_handler(self, level=logging.INFO):
        self.add_handler(logging.StreamHandler(stdout), level=level)

    def set_log_level(self, level):
        self._log_level = level
        for h in self.handlers:
            h.setLevel(self._log_level)
        for name in self.loggers:
            self.loggers[name].setLevel(self._log_level)

    def set_formatter(self, formatter):
        """
        Set all handlers to use formatter
        :param logging.Formatter formatter:
        """
        self.__dict__['formatter'] = formatter
        for h in self.handlers:
            h.setFormatter(self.formatter)

    def configure_handlers_from_config(self):
        configurator = logging.config.BaseConfigurator({})
        handler_classes = {
            'stream_handlers': logging.StreamHandler,
            'file_handlers': logging.FileHandler,
            'timed_rotating_file_handlers': logging.handlers.TimedRotatingFileHandler
        }

        for handler_type in handler_classes:
            for handler_cfg in self.cfg.get(handler_type, []):
                level = logging.getLevelName(handler_cfg.pop('level', self._log_level))

                if 'stream' in handler_cfg:
                    handler_cfg['stream'] = configurator.convert(handler_cfg['stream'])
                handler = handler_classes[handler_type](**handler_cfg)
                self.add_handler(handler, level)


logging_default = LoggingConfiguration(cfg)


class AppLogger:
    """
    Mixin class for logging. An object subclassing this can log using its class name. Contains a
    logging.Logger object and exposes its log methods.
    """
    log_cfg = logging_default

    def debug(self, msg, *args):
        self._logger.debug(msg, *args)

    def info(self, msg, *args):
        self._logger.info(msg, *args)

    def warning(self, msg, *args):
        self._logger.warning(msg, *args)

    def error(self, msg, *args):
        self._logger.error(msg, *args)

    def critical(self, msg, *args):
        self._logger.critical(msg, *args)

    @cached_property
    def _logger(self):
        return self.log_cfg.get_logger(self.__class__.__name__)



import json
import os.path
import unittest
from unittest.mock import Mock


class FakeRestResponse(Mock):
    def __init__(self, *args, **kwargs):
        content = kwargs.pop('content', None)
        if type(content) in (list, dict):
            content = json.dumps(content)
        content = content.encode()
        super().__init__(*args, **kwargs)
        self.content = content
        self.request = Mock(method='a method', path_url='a url')
        self.status_code = 200
        self.reason = 'a reason'

    def json(self):
        return json.loads(self.content.decode('utf-8'))

    @property
    def text(self):
        return self.content.decode('utf-8')


class TestEGCG(unittest.TestCase):
    file_path = os.path.dirname(__file__)
    assets_path = os.path.join(file_path, 'assets')

    etc = os.path.join(os.path.dirname(file_path), 'etc')
    etc_config = os.path.join(etc, 'example_egcg.yaml')

    @staticmethod
    def compare_lists(observed, expected):
        if sorted(observed) != sorted(expected):
            print('')
            print('observed')
            print(observed)
            print('expected')
            print(expected)
            raise AssertionError



import os
from unittest.mock import patch
from egcg_core import archive_management as am
from tests import TestEGCG


class TestArchiveManagement(TestEGCG):
    @patch('egcg_core.archive_management._get_stdout')
    def test_archive_states(self, mocked_stdout):

        mocked_stdout.return_value = 'testfile: (0x0000000d) released exists archived, archive_id:1'
        assert am.archive_states('testfile') == ['released', 'exists', 'archived']

        mocked_stdout.return_value = 'testfile: (0x00000009) exists archived, archive_id:1'
        assert am.archive_states('testfile') == ['exists', 'archived']

        mocked_stdout.return_value = 'testfile: (0x00000001) exists, archive_id:1'
        assert am.archive_states('testfile') == ['exists']

        mocked_stdout.return_value = 'testfile: (0x00000000)'
        assert am.archive_states('testfile') == []

        mocked_stdout.return_value = 'some invalid stdout'
        with self.assertRaises(am.ArchivingError) as c:
            am.archive_states('testfile')
        assert str(c.exception) == 'Could not hsm_state file testfile'

    @patch('egcg_core.archive_management._get_stdout')
    def test_release_file_from_lustre(self, mocked_stdout):
        mocked_stdout.side_effect = [
            'testfile: (0x00000009) exists archived, archive_id:1',
            '',
            'testfile: (0x0000000d) released exists archived, archive_id:1'
        ]
        assert am.release_file_from_lustre('testfile')
        assert mocked_stdout.call_count == 3
        assert mocked_stdout.call_args_list[1][0] == ('lfs hsm_release testfile',)
        mocked_stdout.reset_mock()

        mocked_stdout.side_effect = ['testfile: (0x0000000d) released exists archived, archive_id:1']
        assert am.release_file_from_lustre('testfile')
        assert mocked_stdout.call_count == 1
        mocked_stdout.reset_mock()

        mocked_stdout.side_effect = ['testfile: (0x00000009) exists, archive_id:1']
        with self.assertRaises(am.ArchivingError) as c:
            am.release_file_from_lustre('testfile')
        assert mocked_stdout.call_count == 1
        assert str(c.exception) == 'Cannot release testfile from Lustre because it is not archived to tape'

    @patch('egcg_core.archive_management._get_stdout')
    def test_register_for_archiving(self, mocked_stdout):
        mocked_stdout.side_effect = ['testfile: (0x00000001)', '', 'testfile: (0x00000009) exists, archive_id:1']
        assert am.register_for_archiving('testfile')
        assert mocked_stdout.call_count == 3
        assert mocked_stdout.call_args_list[1][0] == ('lfs hsm_archive testfile',)
        mocked_stdout.reset_mock()

        mocked_stdout.side_effect = ['testfile: (0x00000001) exists, archive_id:1']
        assert am.register_for_archiving('testfile')
        assert mocked_stdout.call_count == 1
        mocked_stdout.reset_mock()

        mocked_stdout.side_effect = [
            'testfile: (0x00000001)', '', 'testfile: (0x00000001)', 'testfile: (0x00000001)', None
        ]
        with self.assertRaises(am.ArchivingError) as c:
            am.register_for_archiving('testfile')
        assert mocked_stdout.call_count == 5
        assert mocked_stdout.call_args_list[1][0] == ('lfs hsm_archive testfile',)
        assert str(c.exception) == 'Registering testfile for archiving to tape failed'

    @patch('egcg_core.archive_management._get_stdout')
    def test_recall_from_tape(self, mocked_stdout):
        mocked_stdout.side_effect = ['testfile: (0x0000000d) released exists archived, archive_id:1', '']
        assert am.recall_from_tape('testfile')
        assert mocked_stdout.call_count == 2
        assert mocked_stdout.call_args_list[1][0] == ('lfs hsm_restore testfile',)

    @patch('egcg_core.archive_management.register_for_archiving')
    def test_archive_directory(self, mocked_register):
        assert am.archive_directory(os.path.join(self.assets_path, 'fastqs'))
        assert mocked_register.call_count == 6



import os
import json
import pytest
from unittest.mock import patch
from tests import FakeRestResponse, TestEGCG
from egcg_core import rest_communication
from egcg_core.util import check_if_nested
from egcg_core.exceptions import RestCommunicationError


def rest_url(endpoint):
    return 'http://localhost:4999/api/0.1/' + endpoint + '/'


def ppath(extension):
    return 'egcg_core.rest_communication.Communicator.' + extension


test_endpoint = 'an_endpoint'
test_nested_request_content = {'data': ['some', {'test': 'content'}]}
test_flat_request_content = {'key1': 'value1', 'key2': 'value2'}
test_patch_document = {
    '_id': '1337', '_etag': 1234567, 'uid': 'a_unique_id', 'list_to_update': ['this', 'that', 'other']
}


def fake_request(method, url, **kwargs):
    if kwargs.get('files'):
        if 'json' in kwargs:
            raise Exception
        if 'data' in kwargs and check_if_nested(kwargs['data']):
            raise Exception
    return FakeRestResponse(status_code=200, content=test_nested_request_content)


patched_response = patch(
    'requests.request',
    side_effect=fake_request
)


auth = ('a_user', 'a_password')


class TestRestCommunication(TestEGCG):
    def setUp(self):
        self.comm = rest_communication.Communicator(auth=auth, baseurl='http://localhost:4999/api/0.1')

    def test_api_url(self):
        assert self.comm.api_url('an_endpoint') == rest_url('an_endpoint')

    def test_parse_query_string(self):
        query_string = 'http://a_url?this=that&other={"another":"more"}&things=1'
        dodgy_query_string = 'http://a_url?this=that?other=another'

        assert self.comm._parse_query_string('http://a_url') == {}
        assert self.comm._parse_query_string(query_string) == {
            'this': 'that', 'other': '{"another":"more"}', 'things': '1'
        }

        with pytest.raises(RestCommunicationError) as e:
            self.comm._parse_query_string(dodgy_query_string)
            assert str(e) == 'Bad query string: ' + dodgy_query_string

        with pytest.raises(RestCommunicationError) as e2:
            self.comm._parse_query_string(query_string, requires=['thangs'])
            assert str(e2) == query_string + " did not contain all required fields: ['thangs']"

    def test_detect_files_in_json(self):
        json_no_files = {'k1': 'v1', 'k2': 'v2'}
        obs_files, obs_json = self.comm._detect_files_in_json(json_no_files)
        file_path = os.path.join(self.assets_path, 'test_to_upload.txt')
        assert obs_files is None
        assert obs_json == json_no_files

        json_with_files = {'k1': 'v1', 'k2': ('file', file_path)}
        obs_files, obs_json = self.comm._detect_files_in_json(json_with_files)
        assert obs_files == {'k2': (file_path, b'test content', 'text/plain')}
        assert obs_json == {'k1': 'v1'}

        json_list = [json_with_files, json_with_files]
        obs_files, obs_json = self.comm._detect_files_in_json(json_list)
        assert obs_files == [
            {'k2': (file_path, b'test content', 'text/plain')},
            {'k2': (file_path, b'test content', 'text/plain')}
        ]
        assert obs_json == [{'k1': 'v1'}, {'k1': 'v1'}]

    @patched_response
    def test_req(self, mocked_response):
        json_content = ['some', {'test': 'json'}]
        response = self.comm._req('METHOD', rest_url(test_endpoint), json=json_content)
        assert response.status_code == 200
        assert json.loads(response.content.decode('utf-8')) == response.json() == test_nested_request_content
        mocked_response.assert_called_with('METHOD', rest_url(test_endpoint), auth=auth, json=json_content)

    def test_get_documents_depaginate(self):
        docs = (
            FakeRestResponse(content={'data': ['this', 'that'], '_links': {'next': {'href': 'an_endpoint?max_results=101&page=2'}}}),
            FakeRestResponse(content={'data': ['other', 'another'], '_links': {'next': {'href': 'an_endpoint?max_results=101&page=3'}}}),
            FakeRestResponse(content={'data': ['more', 'things'], '_links': {}})
        )
        with patch(ppath('_req'), side_effect=docs) as mocked_req:
            assert self.comm.get_documents('an_endpoint', all_pages=True, max_results=101) == [
                'this', 'that', 'other', 'another', 'more', 'things'
            ]
            assert all([a[0][1].startswith(rest_url('an_endpoint')) for a in mocked_req.call_args_list])
            assert [a[1] for a in mocked_req.call_args_list] == [
                # Communicator.get_content passes ints
                {'params': {'page': 1, 'max_results': 101}, 'quiet': False},
                # url parsing passes strings, but requests removes the quotes anyway
                {'params': {'page': '2', 'max_results': '101'}, 'quiet': False},
                {'params': {'page': '3', 'max_results': '101'}, 'quiet': False}
            ]

        docs = [
            FakeRestResponse(
                content={
                    'data': ['data%s' % d],
                    '_links': {'next': {'href': 'an_endpoint?max_results=101&page=%s' % d}}
                }
            )
            for d in range(1, 1200)
        ]
        docs.append(FakeRestResponse(content={'data': ['last piece'], '_links': {}}))

        with patch(ppath('_req'), side_effect=docs):
            ret = self.comm.get_documents('an_endpoint', all_pages=True, max_results=101)
            assert len(ret) == 1200

    @patched_response
    def test_get_content(self, mocked_response):
        data = self.comm.get_content(test_endpoint, max_results=100, where={'a_field': 'thing'})
        assert data == test_nested_request_content
        assert mocked_response.call_args[0][1].startswith(rest_url(test_endpoint))
        assert mocked_response.call_args[1] == {
            'auth': ('a_user', 'a_password'),
            'params': {'max_results': 100, 'where': '{"a_field": "thing"}', 'page': 1}
        }

    def test_get_documents(self):
        with patched_response:
            data = self.comm.get_documents(test_endpoint, max_results=100, where={'a_field': 'thing'})
            assert data == test_nested_request_content['data']

    def test_get_document(self):
        expected = test_nested_request_content['data'][0]
        with patched_response:
            observed = self.comm.get_document(test_endpoint, max_results=100, where={'a_field': 'thing'})
            assert observed == expected

    @patched_response
    def test_post_entry(self, mocked_response):
        self.comm.post_entry(test_endpoint, payload=test_nested_request_content)
        mocked_response.assert_called_with(
            'POST',
            rest_url(test_endpoint),
            auth=auth,
            json=test_nested_request_content,
            files=None
        )
        file_path = os.path.join(self.assets_path, 'test_to_upload.txt')
        test_request_content_plus_files = dict(test_flat_request_content)
        test_request_content_plus_files['f'] = ('file', file_path)
        self.comm.post_entry(test_endpoint, payload=test_request_content_plus_files)
        mocked_response.assert_called_with(
            'POST',
            rest_url(test_endpoint),
            auth=auth,
            data=test_flat_request_content,
            files={'f': (file_path, b'test content', 'text/plain')}
        )

        self.comm.post_entry(test_endpoint, payload=test_flat_request_content, use_data=True)
        mocked_response.assert_called_with(
            'POST',
            rest_url(test_endpoint),
            auth=auth,
            data=test_flat_request_content,
            files=None
        )

        self.comm.post_entry(test_endpoint, payload=test_request_content_plus_files, use_data=True)
        mocked_response.assert_called_with(
            'POST',
            rest_url(test_endpoint),
            auth=auth,
            data=test_flat_request_content,
            files={'f': (file_path, b'test content', 'text/plain')}
        )

    @patched_response
    def test_put_entry(self, mocked_response):
        self.comm.put_entry(test_endpoint, 'an_element_id', payload=test_nested_request_content)
        mocked_response.assert_called_with(
            'PUT',
            rest_url(test_endpoint) + 'an_element_id',
            auth=auth,
            json=test_nested_request_content,
            files=None
        )

        file_path = os.path.join(self.assets_path, 'test_to_upload.txt')
        test_request_content_plus_files = dict(test_flat_request_content)
        test_request_content_plus_files['f'] = ('file', file_path)
        self.comm.put_entry(test_endpoint, 'an_element_id', payload=test_request_content_plus_files)
        mocked_response.assert_called_with(
            'PUT',
            rest_url(test_endpoint) + 'an_element_id',
            auth=auth,
            data=test_flat_request_content,
            files={'f': (file_path, b'test content', 'text/plain')}
        )

    @patch(ppath('get_document'), return_value=test_patch_document)
    @patched_response
    def test_patch_entry(self, mocked_response, mocked_get_doc):
        patching_payload = {'list_to_update': ['another']}
        self.comm.patch_entry(
            test_endpoint,
            payload=patching_payload,
            id_field='uid',
            element_id='a_unique_id',
            update_lists=['list_to_update']
        )

        mocked_get_doc.assert_called_with(test_endpoint, where={'uid': 'a_unique_id'})
        mocked_response.assert_called_with(
            'PATCH',
            rest_url(test_endpoint) + '1337',
            headers={'If-Match': 1234567},
            auth=auth,
            json={'list_to_update': ['this', 'that', 'other', 'another']},
            files=None
        )

    @patch(ppath('get_document'), return_value=test_patch_document)
    @patched_response
    def test_auth_token_and_if_match(self, mocked_response, mocked_get_doc):
        self.comm._auth = 'an_auth_token'
        self.comm.patch_entry(test_endpoint, {'this': 'that'}, 'uid', 'a_unique_id')
        mocked_get_doc.assert_called_with(test_endpoint, where={'uid': 'a_unique_id'})
        mocked_response.assert_called_with(
            'PATCH',
            rest_url(test_endpoint) + '1337',
            headers={'If-Match': 1234567, 'Authorization': 'Token an_auth_token'},
            json={'this': 'that'},
            files=None
        )

    def test_post_or_patch(self):
        test_post_or_patch_payload = {'uid': '1337', 'list_to_update': ['more'], 'another_field': 'that'}
        test_post_or_patch_payload_no_uid = {'list_to_update': ['more'], 'another_field': 'that'}
        test_post_or_patch_doc = {
            'uid': 'a_uid', '_id': '1337', '_etag': 1234567, 'list_to_update': ['things'], 'another_field': 'this'
        }
        patched_post = patch(ppath('post_entry'), return_value=True)
        patched_patch = patch(ppath('_patch_entry'), return_value=True)
        patched_get = patch(ppath('get_document'), return_value=test_post_or_patch_doc)
        patched_get_none = patch(ppath('get_document'), return_value=None)

        with patched_get as mget, patched_patch as mpatch:
            self.comm.post_or_patch(
                'an_endpoint',
                [test_post_or_patch_payload],
                id_field='uid',
                update_lists=['list_to_update']
            )
            mget.assert_called_with('an_endpoint', where={'uid': '1337'})
            mpatch.assert_called_with(
                'an_endpoint',
                test_post_or_patch_doc,
                test_post_or_patch_payload_no_uid,
                ['list_to_update']
            )

        with patched_get_none as mget, patched_post as mpost:
            self.comm.post_or_patch(
                'an_endpoint', [test_post_or_patch_payload], id_field='uid', update_lists=['list_to_update']
            )
            mget.assert_called_with('an_endpoint', where={'uid': '1337'})
            mpost.assert_called_with('an_endpoint', test_post_or_patch_payload)

    def test_token_auth(self):
        hashed_token = '{"some": "hashed"}.tokenauthentication'
        self.comm._auth = hashed_token
        with patched_response as p:
            self.comm._req('GET', self.comm.baseurl + 'an_endpoint')
            p.assert_called_with(
                'GET',
                self.comm.baseurl + 'an_endpoint',
                headers={'Authorization': 'Token ' + hashed_token}
            )


def test_default():
    d = rest_communication.default
    assert d.baseurl == 'http://localhost:4999/api/0.1'
    assert d.auth == ('a_user', 'a_password')



from egcg_core.app_logging import AppLogger
from egcg_core.config import cfg
from .asana import AsanaNotification
from .email import EmailNotification, send_email
from .log import LogNotification


class NotificationCentre(AppLogger):
    ntf_aliases = {
        'log': LogNotification,
        'email': EmailNotification,
        'asana': AsanaNotification
    }

    def __init__(self, name):
        self.name = name
        self.subscribers = {}

        for s in cfg.get('notifications', {}):
            if s in self.ntf_aliases:
                self.debug('Configuring notification for: ' + s)
                config = cfg['notifications'][s]
                self.subscribers[s] = self.ntf_aliases[s](name=self.name, **config)
            else:
                self.warning("Bad notification config '%s' - this will be ignored" % s)

    def notify(self, msg, subs):
        for s in subs:
            if s in self.subscribers:
                self.subscribers[s].notify(msg)
            else:
                self.debug('Tried to notify by %s, but no configuration present', s)

    def notify_all(self, msg):
        for name, s in self.subscribers.items():
            s.notify(msg)



import os
import re
import subprocess
from time import sleep
from egcg_core.app_logging import logging_default as log_cfg
from egcg_core.exceptions import ArchivingError

app_logger = log_cfg.get_logger('archive_management')
state_re = re.compile('^(.+): \((0x\w+)\)(.+)?')


def _get_stdout(cmd):
    p = subprocess.Popen(cmd.split(' '), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    exit_status = p.wait()
    o, e = p.stdout.read(), p.stderr.read()
    msg = '%s -> (%s, %s, %s)' % (cmd, exit_status, o, e)
    if exit_status:
        app_logger.error(msg)
        return None
    else:
        app_logger.debug(msg)
        return o.decode('utf-8').strip()


def archive_states(file_path):
    val = _get_stdout('lfs hsm_state ' + file_path)
    match = state_re.match(val)
    if match:
        file_name = match.group(1)
        assert file_name == file_path
        state_and_id = match.group(3)
        if state_and_id:
            state, archive_id = state_and_id.split(',')
            states = state.strip().split()
            return states
        else:
            return []
    else:
        raise ArchivingError('Could not hsm_state file %s' % file_path)


def is_of_state(state, file_path, known_states=None):
    if known_states:
        return state in known_states
    else:
        return state in archive_states(file_path)


def is_registered_for_archiving(file_path, known_states=None):
    return is_of_state('exists', file_path, known_states)


def is_archived(file_path, known_states=None):
    return is_of_state('archived', file_path, known_states)


def is_released(file_path, known_states=None):
    return is_of_state('released', file_path, known_states)


def is_dirty(file_path, known_states=None):
    return is_of_state('dirty', file_path, known_states)


def release_file_from_lustre(file_path):
    states = archive_states(file_path)  # store the states to avoid querying multiple times
    if is_dirty(file_path, states):
        raise ArchivingError('File %s is in a dirty state' % file_path)
    if not is_archived(file_path, states):
        raise ArchivingError('Cannot release %s from Lustre because it is not archived to tape' % file_path)

    if not is_released(file_path, states):
        val = _get_stdout('lfs hsm_release ' + file_path)
        if val is not None:
            return is_released(file_path)
    else:
        app_logger.debug('Trying to release a %s already released from Lustre', file_path)
        return True


def register_for_archiving(file_path, strict=False):
    if is_registered_for_archiving(file_path):
        return True

    val = _get_stdout('lfs hsm_archive ' + file_path)
    if val is None or not is_registered_for_archiving(file_path):
        if strict:
            raise ArchivingError('Registering %s for archiving to tape failed' % file_path)
        sleep(1)  # registering can sometimes take time, so give it a second...
        return register_for_archiving(file_path, strict=True)  # ... and try again (once only)

    return True


def recall_from_tape(file_path):
    states = archive_states(file_path)
    if is_dirty(file_path, states):
        raise ArchivingError('File %s is in a dirty state' % file_path)

    if is_archived(file_path, states) and is_released(file_path, states):
        val = _get_stdout('lfs hsm_restore ' + file_path)
        if val is not None:
            return True


def archive_directory(directory):
    """Recursively archive all the files in a directory"""
    success = True
    for f in os.listdir(directory):
        fp = os.path.join(directory, f)
        if os.path.isdir(fp):
            success = success and archive_directory(fp)
        elif os.path.isfile(fp):
            success = success and register_for_archiving(fp)
    return success



import os
import sys
import logging
import logging.handlers
from unittest.mock import Mock
from tests import TestEGCG
from egcg_core import app_logging
from egcg_core.config import cfg
cfg.load_config_file(TestEGCG.etc_config)


class TestLoggingConfiguration(TestEGCG):
    def setUp(self):
        self.log_cfg = app_logging.LoggingConfiguration(cfg['logging'])

    def tearDown(self):
        self.log_cfg = None

    def test_formatters(self):
        default = self.log_cfg.default_formatter
        assert default.datefmt == '%Y-%b-%d %H:%M:%S'
        assert default._fmt == '[%(asctime)s][%(name)s][%(levelname)s] %(message)s'

        blank = self.log_cfg.blank_formatter
        assert blank.datefmt is None
        assert blank._fmt == '%(message)s'

        assert self.log_cfg.formatter is default

        assert self.log_cfg.handlers == set()
        assert self.log_cfg._log_level == logging.INFO

    def test_get_logger(self):
        l = self.log_cfg.get_logger('a_logger')
        assert l.level == self.log_cfg._log_level
        assert l in self.log_cfg.loggers.values()
        assert list(self.log_cfg.handlers) == l.handlers

    def test_add_handler(self):
        h = logging.StreamHandler(stream=sys.stdout)
        self.log_cfg.add_handler(h)

        assert h in self.log_cfg.handlers
        assert h.formatter is self.log_cfg.formatter
        assert h.level == logging.INFO

    def test_set_log_level(self):
        h = logging.StreamHandler(stream=sys.stdout)
        self.log_cfg.add_handler(h, level=logging.INFO)
        l = self.log_cfg.get_logger('a_logger')
        assert h.level == l.level == logging.INFO

        self.log_cfg.set_log_level(logging.DEBUG)
        assert h.level == l.level == logging.DEBUG

    def test_set_formatter(self):
        h = logging.StreamHandler(stream=sys.stdout)
        self.log_cfg.add_handler(h)

        self.log_cfg.set_formatter(self.log_cfg.blank_formatter)
        assert h.formatter is self.log_cfg.blank_formatter
        self.log_cfg.set_formatter(self.log_cfg.default_formatter)
        assert h.formatter is self.log_cfg.default_formatter

    def test_configure_handlers_from_config(self):
        test_log = os.path.join(self.assets_path, 'test.log')
        self.log_cfg.configure_handlers_from_config()
        for h in self.log_cfg.handlers:
            if type(h) is logging.StreamHandler:
                assert h.stream is sys.stdout and h.level == logging.DEBUG
            elif type(h) is logging.FileHandler:
                assert h.stream.name == test_log and h.level == logging.WARNING
            elif type(h) is logging.handlers.TimedRotatingFileHandler:
                assert h.stream.name == test_log and h.level == logging.INFO
                assert h.when == 'H' and h.interval == 3600  # casts 'h' to 'H' and multiplies when to seconds


class TestAppLogger(TestEGCG):
    def setUp(self):
        self.log_cfg = app_logging.LoggingConfiguration({})
        self.app_logger = app_logging.AppLogger()
        self.app_logger.log_cfg = self.log_cfg

    def tearDown(self):
        self.app_logger = None
        self.log_cfg = None

    def test_get_logger(self):
        self.log_cfg.get_logger = Mock()
        self.app_logger.info('Things')
        self.app_logger._logger.info.assert_called_with('Things')
        self.log_cfg.get_logger.assert_called_with('AppLogger')



import json
import mimetypes
import requests
from urllib.parse import urljoin
from egcg_core.config import cfg
from egcg_core.app_logging import AppLogger
from egcg_core.exceptions import RestCommunicationError
from egcg_core.util import check_if_nested


class Communicator(AppLogger):
    successful_statuses = (200, 201, 202, 204)

    def __init__(self, auth=None, baseurl=None):
        self._baseurl = baseurl
        self._auth = auth

    @staticmethod
    def serialise(queries):
        serialised_queries = {}
        for k, v in queries.items():
            serialised_queries[k] = json.dumps(v) if type(v) is dict else v
        return serialised_queries

    @staticmethod
    def get_file_content(file_name):
        """
        This method open in binary mode th file provided.
        :param file_name: path to the file
        :return: file content
        """
        with open(file_name, 'rb') as f:
            fc = f.read()
        return fc

    @staticmethod
    def _extract_files(json_dict):
        """
        Take a json dict and search all values for a tuple['file',<file_path>]. Extract file_path, read its content and
        create another tuple ready to be used by requests. Also modify the original json input to remove the file entry.
        :param dict json_dict:
        :return: a dict of: {field to update: (filename, file content, file MIME type)}
        """
        files = {}
        for k in json_dict:
            if isinstance(json_dict[k], tuple) and len(json_dict[k]) > 1 and json_dict[k][0] == 'file':
                # Construct the 3 parts file upload tuple as requests requires it
                files[k] = (
                    json_dict[k][1],                                 # file name
                    Communicator.get_file_content(json_dict[k][1]),  # file content
                    mimetypes.guess_type(json_dict[k][1])[0]         # file mime type
                )
        if files:
            # remove the extracted files from the original json
            for f in files:
                json_dict.pop(f)
            return files
        return None

    @staticmethod
    def _detect_files_in_json(json_data):
        if isinstance(json_data, list):
            # Assume a list of dict
            list_files = []
            list_jsons = []
            for d in json_data:
                _d = dict(d)
                list_files.append(Communicator._extract_files(_d))
                list_jsons.append(_d)
            return list_files, list_jsons
        if isinstance(json_data, dict):
            # Assume a single dict
            _json_dict = dict(json_data)
            return Communicator._extract_files(_json_dict), _json_dict

    @property
    def baseurl(self):
        if self._baseurl is None:
            self._baseurl = cfg['rest_api']['url'].rstrip('/')
        return self._baseurl

    @property
    def auth(self):
        if self._auth is None and 'username' in cfg['rest_api'] and 'password' in cfg['rest_api']:
            self._auth = (cfg['rest_api']['username'], cfg['rest_api']['password'])
        return self._auth

    def api_url(self, endpoint):
        return '{base_url}/{endpoint}/'.format(base_url=self.baseurl, endpoint=endpoint)

    @staticmethod
    def _parse_query_string(query_string, requires=None):
        if '?' not in query_string:
            return {}

        if query_string.count('?') != 1:
            raise RestCommunicationError('Bad query string: ' + query_string)

        href, query = query_string.split('?')
        query = dict(x.split('=') for x in query.split('&'))
        if requires and not all(r in query for r in requires):
            raise RestCommunicationError('%s did not contain all required fields: %s' % (query_string, requires))
        return query

    def _req(self, method, url, quiet=False, **kwargs):
        if type(self.auth) is tuple:
            kwargs['auth'] = self.auth
        elif type(self.auth) is str:
            kwargs['headers'] = dict(kwargs.get('headers', {}), Authorization='Token %s' % self.auth)

        # can't upload json and files at the same time, so we need to move the json parameter to data
        # data can't upload complex structures that would require json encoding.
        # this means we can't upload data with nested lists/dicts at the same time as files
        if kwargs.get('files') and kwargs.get('json'):
            if check_if_nested(kwargs.get('json')):
                raise RestCommunicationError('Cannot upload files and nested json in one query')
            kwargs['data'] = kwargs.pop('json')

        r = requests.request(method, url, **kwargs)

        kwargs.pop('auth', None)
        kwargs.pop('headers', None)
        kwargs.pop('files', None)
        # e.g: 'POST <url> ({"some": "args"}) -> {"some": "content"}. Status code 201. Reason: CREATED
        report = '%s %s (%s) -> %s. Status code %s. Reason: %s' % (
            r.request.method, r.request.path_url, kwargs, r.content.decode('utf-8'), r.status_code, r.reason
        )
        if r.status_code in self.successful_statuses:
            if not quiet:
                self.debug(report)
        else:
            self.error(report)
            raise RestCommunicationError('Encountered a %s status code: %s' % (r.status_code, r.reason))
        return r

    def get_content(self, endpoint, paginate=True, quiet=False, **query_args):
        if paginate:
            query_args['max_results'] = query_args.pop('max_results', 100)  # default to page size of 100
            query_args['page'] = query_args.pop('page', 1)

        url = self.api_url(endpoint)
        return self._req('GET', url, quiet=quiet, params=self.serialise(query_args)).json()

    def get_documents(self, endpoint, paginate=True, all_pages=False, quiet=False, **query_args):
        content = self.get_content(endpoint, paginate, quiet, **query_args)
        elements = content['data']

        if all_pages:
            while 'next' in content['_links']:
                next_query = self._parse_query_string(content['_links']['next']['href'], requires=('max_results', 'page'))
                query_args.update(next_query)
                content = self.get_content(endpoint, paginate, quiet, **query_args)
                elements.extend(content['data'])

        return elements

    def get_document(self, endpoint, idx=0, **query_args):
        documents = self.get_documents(endpoint, **query_args)
        if documents:
            return documents[idx]
        else:
            self.warning('No document found in endpoint %s for %s', endpoint, query_args)

    def post_entry(self, endpoint, payload, use_data=False):
        files, payload = self._detect_files_in_json(payload)
        if use_data:
            return self._req('POST', self.api_url(endpoint), data=payload, files=files)
        else:
            return self._req('POST', self.api_url(endpoint), json=payload, files=files)

    def put_entry(self, endpoint, element_id, payload):
        files, payload = self._detect_files_in_json(payload)
        return self._req('PUT', urljoin(self.api_url(endpoint), element_id), json=payload, files=files)

    def _patch_entry(self, endpoint, doc, payload, update_lists=None):
        """
        Patch a specific database item (specified by doc) with the given data payload.
        :param str endpoint:
        :param dict doc: The entry in the database to patch (contains the relevant _id and _etag)
        :param dict payload: Data with which to patch doc
        :param list update_lists: Doc items listed here will be appended rather than replaced by the patch
        """
        url = urljoin(self.api_url(endpoint), doc['_id'])
        files, payload = self._detect_files_in_json(payload)
        _payload = dict(payload)
        if update_lists:
            for l in update_lists:
                content = doc.get(l, [])
                new_content = [x for x in _payload.get(l, []) if x not in content]
                _payload[l] = content + new_content
        return self._req('PATCH', url, headers={'If-Match': doc['_etag']}, json=_payload, files=files)

    def patch_entry(self, endpoint, payload, id_field, element_id, update_lists=None):
        """
        Retrieve a document at the given endpoint with the given unique ID, and patch it with some data.
        :param str endpoint:
        :param dict payload:
        :param str id_field: The name of the unique identifier (e.g. 'run_element_id', 'proc_id', etc.)
        :param element_id: The value of id_field to retrieve (e.g. '160301_2_ATGCATGC')
        :param list update_lists:
        """
        doc = self.get_document(endpoint, where={id_field: element_id})
        if doc:
            return self._patch_entry(endpoint, doc, payload, update_lists)

    def patch_entries(self, endpoint, payload, update_lists=None, **query_args):
        """
        Retrieve many documents and patch them all with the same data.
        :param str endpoint:
        :param dict payload:
        :param list update_lists:
        :param query_args: Database query args to pass to get_documents
        """
        docs = self.get_documents(endpoint, **query_args)
        if docs:
            self.info('Updating %s docs matching %s', len(docs), query_args)
            for doc in docs:
                self._patch_entry(endpoint, doc, payload, update_lists)

    def post_or_patch(self, endpoint, input_json, id_field=None, update_lists=None):
        """
        For each document supplied, either post to the endpoint if the unique id doesn't yet exist there, or
        patch if it does.
        :param str endpoint:
        :param input_json: A list of documents to post or patch to the endpoint.
        :param str id_field: The field to use as the unique ID for the endpoint.
        :param list update_lists:
        """
        for payload in input_json:
            _payload = dict(payload)
            doc = self.get_document(endpoint, where={id_field: _payload[id_field]})
            if doc:
                _payload.pop(id_field)
                self._patch_entry(endpoint, doc, _payload, update_lists)
            else:
                self.post_entry(endpoint, _payload)


default = Communicator()
api_url = default.api_url
get_content = default.get_content
get_documents = default.get_documents
get_document = default.get_document
post_entry = default.post_entry
put_entry = default.put_entry
patch_entry = default.patch_entry
patch_entries = default.patch_entries
post_or_patch = default.post_or_patch



import subprocess
from time import sleep
from egcg_core.exceptions import EGCGError
from egcg_core.app_logging import AppLogger
from egcg_core.config import cfg
from . import script_writers

running_executors = {}


def stop_running_jobs():
    for job_id in running_executors:
        running_executors[job_id].cancel_job()

    for job_id in list(running_executors):
        running_executors[job_id].join()


class ClusterExecutor(AppLogger):
    script_writer = script_writers.ScriptWriter
    finished_statuses = None
    unfinished_statuses = None

    def __init__(self, *cmds, prelim_cmds=None, **cluster_config):
        """
        :param cmds: Full path to a job submission script
        """
        self.interval = cfg.query('executor', 'join_interval', ret_default=30)
        self.job_id = None
        self.job_name = cluster_config.get('job_name')
        self.cmds = cmds
        self.prelim_cmds = prelim_cmds
        self.writer = self._get_writer(job_queue=cfg['executor']['job_queue'], **cluster_config)

    def write_script(self):
        if self.prelim_cmds:
            self.writer.register_cmds(*self.prelim_cmds, parallel=False)

        pre_job_source = cfg.query('executor', 'pre_job_source')
        if pre_job_source:
            self.writer.register_cmd('source ' + pre_job_source)

        self.writer.line_break()
        self.writer.register_cmds(*self.cmds, parallel=True)
        self.writer.add_header()
        self.writer.save()

    def start(self):
        """Write the jobs into a script, submit it and capture qsub's output as self.job_id."""
        self.write_script()
        self._submit_job()
        running_executors[self.job_id] = self  # register to running_executors
        self.info('Submitted "%s" as job %s' % (self.writer.script_name, self.job_id))

    def join(self):
        """Wait until the job has finished, then return its exit status."""
        sleep(5)
        while not self._job_finished():
            sleep(self.interval)
        running_executors.pop(self.job_id, None)  # unregister from running_executors
        return self._job_exit_code()

    def _get_writer(self, job_name, working_dir, job_queue, walltime=None, cpus=1, mem=2, log_commands=True):
        return self.script_writer(job_name, working_dir, job_queue, log_commands=log_commands, cpus=cpus, mem=mem, walltime=walltime)

    def _job_statuses(self):
        return ()

    def _job_exit_code(self):
        raise NotImplementedError

    def _submit_job(self):
        self.job_id = self._run_and_retry(cfg['executor']['qsub'] + ' ' + self.writer.script_name)
        if self.job_id is None:
            raise EGCGError('Job submission failed')

    def _job_finished(self):
        statuses = self._job_statuses()
        for s in statuses:
            if s in self.finished_statuses:
                pass
            elif s in self.unfinished_statuses:
                return False
            else:
                raise EGCGError('Bad job status: %s', s)
        return True

    def _get_stdout(self, cmd):
        p = subprocess.Popen(cmd.split(' '), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        exit_status = p.wait()
        o, e = p.stdout.read(), p.stderr.read()
        self.debug('%s -> (%s, %s, %s)', cmd, exit_status, o, e)
        if exit_status:
            return None
        else:
            return o.decode('utf-8').strip()

    def _run_and_retry(self, cmd, retry=3):
        attempt = 0
        while attempt < retry:
            msg = self._get_stdout(cmd)
            if msg is not None:
                return msg
            sleep(5)
            attempt += 1

    def cancel_job(self):
        if not self._job_finished():
            self._cancel_job()

    def _cancel_job(self):
        raise NotImplementedError


class PBSExecutor(ClusterExecutor):
    script_writer = script_writers.PBSWriter
    unfinished_statuses = 'BEHQRSTUW'
    finished_statuses = 'FXM'

    def _qstat(self):
        data = self._run_and_retry('qstat -xt {j}'.format(j=self.job_id)).split('\n')
        return [d for d in data[2:] if d]

    def _job_statuses(self):
        statuses = set()
        reports = self._qstat()
        for r in reports:
            job_id, job_name, user, time, status, queue = r.split()
            statuses.add(status)
        return statuses

    def _job_exit_code(self):
        exit_status = 0
        for s in self._job_statuses():
            exit_status += self.finished_statuses.index(s)
        return exit_status

    def _cancel_job(self):
        msg = self._run_and_retry('qdel ' + self.job_id)
        self.info(msg)


class SlurmExecutor(ClusterExecutor):
    script_writer = script_writers.SlurmWriter
    unfinished_statuses = ('CONFIGURING', 'COMPLETING', 'PENDING', 'RUNNING', 'RESIZING', 'SUSPENDED',)
    finished_statuses = ('BOOT_FAIL', 'CANCELLED', 'COMPLETED', 'DEADLINE', 'FAILED', 'NODE_FAIL',
                         'PREEMPTED', 'TIMEOUT')

    def _submit_job(self):
        # sbatch stdout: "Submitted batch job {job_id}"
        super()._submit_job()
        self.job_id = self.job_id.split()[-1].strip()

    def _sacct(self, output_format):
        data = self._run_and_retry('sacct -nX -j {j} -o {o}'.format(j=self.job_id, o=output_format))
        return set(d.strip() for d in data.split('\n'))

    def _squeue(self):
        s = self._run_and_retry('squeue -h -j {j} -o %T'.format(j=self.job_id))
        if s:
            return set(s.split('\n'))

    def _job_statuses(self):
        s = self._squeue()
        if s:  # job is in squeue, so use that
            return s
        return set(s.rstrip('+') for s in self._sacct('State'))  # job no longer in squeue, so use sacct

    def _job_exit_code(self):
        exit_status = 0
        states = set()
        reports = self._sacct('State,ExitCode')
        for r in reports:
            state, exit_code = r.split()
            state = state.rstrip('+')
            states.add(state)
            exit_code = int(exit_code.split(':')[0])
            if state == 'CANCELLED' and not exit_code:  # cancelled jobs can still be exit status 0
                self.debug('Found a cancelled job - using exit status 9')
                exit_code = 9
            exit_status += exit_code

        self.info('Got %s states from %s (%s) with %s jobs: %s', len(states), self.job_name, self.job_id,
                  len(reports), states)
        return exit_status

    def _cancel_job(self):
        msg = self._run_and_retry('scancel ' + self.job_id)
        self.info(msg)



import re
import sys
from os.path import abspath, dirname, basename

table = str.maketrans('', '', '\'" []')


class ConfigRef:
    def __init__(self, ref, lineno):
        self.ref = ref
        self.lineno = lineno

    def __repr__(self):
        return '%s at line %s' % ('/'.join(self.ref), self.lineno)


def find_config_references(file):
    getitems = []
    gets = []
    queries = []
    imports = _config_import_references(file)
    lineno = 0

    with open(file) as f:
        for line in f:
            lineno += 1
            for i in imports:
                for l, m in (
                    (getitems, _find_get_item),
                    (gets, _find_gets),
                    (queries, _find_query)
                ):
                    refs = m(line, i, lineno)
                    if refs:
                        l.extend(refs)

    return getitems, gets, queries


def _find_gets(line, cfg_pointer, lineno):
    return [ConfigRef(x, lineno) for x in re.findall(cfg_pointer + '\.get\([\'"]([a-z]+)[\'"]\)', line)]


def _find_query(line, cfg_pointer, lineno):
    f = re.findall(cfg_pointer + '\.query\(([^\)]+)\)', line)
    if f:
        l = []
        for x in f[0].split(','):
            x = x.translate(table)
            if len(x) > 3 and '.' not in x:
                l.append(x)
        return [ConfigRef(l, lineno)]


def _find_get_item(line, cfg_pointer, lineno):
    f = re.findall(cfg_pointer + '(\[.+\])', line)
    if f:
        l = [x.translate(table) for x in f[0].split(']')]
        return [ConfigRef([x for x in l if x], lineno)]


def _config_import_references(file):
    refs = []
    with open(file) as f:
        for line in f:
            if line.startswith('import') or (line.startswith('from') and 'import' in line):
                if 'config' in line:
                    refs.append(line.strip().split(' ')[-1])
    return refs


def validate_main_config(cfg, file):
    print('_' * 36)
    print('Searching for missing configs in %s:' % basename(file))
    getitems, gets, queries = find_config_references(file)
    for i in getitems + queries:
        if not cfg.query(*i.ref):
            print(i)

    for i in gets:
        if not cfg.get(i):
            print(i)
    print('_' * 36)


def main():
    sys.path.append(dirname(dirname(abspath(__file__))))
    from egcg_core.config import cfg

    config_file = sys.argv[1]
    source_code = sys.argv[2]

    cfg.load_config_file(config_file)
    validate_main_config(cfg, source_code)


if __name__ == '__main__':
    main()



from .executor import Executor
from .stream_executor import StreamExecutor
from .array_executor import ArrayExecutor
from .cluster_executor import PBSExecutor, SlurmExecutor, stop_running_jobs
from egcg_core.config import cfg
from egcg_core.exceptions import EGCGError


def local_execute(*cmds, parallel=True):
    """
    Execute commands locally
    :param cmds:
    :param parallel: Whether to execute multiple cmds in parallel or sequentially
    :return: Executor
    """
    if len(cmds) == 1:
        if parallel:
            e = StreamExecutor(cmds[0])
        else:
            e = Executor(cmds[0])
    else:
        e = ArrayExecutor(cmds, stream=parallel)

    e.start()
    return e


def cluster_execute(*cmds, env=None, prelim_cmds=None, **cluster_config):
    """
    Execute commands on a compute cluster
    :param cmds:
    :param env: The kind of resource manager being run
    :param prelim_cmds: Any commands to execute before starting a job array
    :param cluster_config:
    :return: ClusterExecutor
    """
    if env is None:
        env = cfg.query('executor', 'job_execution')

    if env == 'pbs':
        cls = PBSExecutor
    elif env == 'slurm':
        cls = SlurmExecutor
    else:
        raise EGCGError('Unknown execution environment: %s' % env)

    e = cls(*cmds, prelim_cmds=prelim_cmds, **cluster_config)
    e.start()
    return e


def execute(*cmds, env=None, prelim_cmds=None, **cluster_config):
    if env is None:
        env = cfg.query('executor', 'job_execution')

    if env == 'local':
        return local_execute(*cmds)
    else:
        return cluster_execute(*cmds, env=env, prelim_cmds=prelim_cmds, **cluster_config)



from os.path import join
from egcg_core.app_logging import AppLogger
from egcg_core.exceptions import EGCGError


class ScriptWriter(AppLogger):
    """
    Writes a basic job submission script. Subclassed by PBSWriter. Initialises with self.lines as an empty
    list, which is appended by self.write_line. This list is then saved line by line to self.script_file by
    self.save.
    """
    header = (
        '#!/bin/bash\n',
        '# job name: {job_name}',
        '# cpus: {cpus}',
        '# mem: {mem}gb',
        '# queue: {job_queue}',
        '# log file: {log_file}'
    )
    walltime_header = '# walltime: {walltime}'
    array_header = '# job array: 1-{jobs}'
    suffix = '.sh'
    array_index = 'JOB_INDEX'

    def __init__(self, job_name, working_dir, job_queue, log_commands=True, **cluster_config):
        """
        :param str job_name: Desired full path to the pbs script to write
        """
        self.script_name = join(working_dir, job_name + self.suffix)
        self.log_commands = log_commands
        self.working_dir = working_dir
        self.log_file = join(self.working_dir, job_name + '.log')
        self.info('Writing job "%s" in %s', job_name, working_dir)
        self.lines = []
        self.array_jobs_written = 0
        self.cluster_config = dict(cluster_config, job_name=job_name, log_file=self.log_file, job_queue=job_queue)

    def register_cmd(self, cmd, log_file=None):
        if log_file:
            cmd += ' > %s 2>&1' % log_file
        self.add_line(cmd)

    def register_cmds(self, *cmds, parallel):
        if parallel:
            self.add_job_array(*cmds)
        else:
            self.lines.extend(list(cmds))

    def add_job_array(self, *cmds):
        if self.array_jobs_written != 0:
            raise EGCGError('Already written a job array - can only have one per script')

        if len(cmds) == 1:
            self.register_cmd(cmds[0])
        else:
            self._start_array()
            for idx, cmd in enumerate(cmds):
                self._register_array_cmd(
                    idx + 1,
                    cmd,
                    log_file=self.log_file + str(idx + 1) if self.log_commands else None
                )
            self._finish_array()

        self.array_jobs_written += len(cmds)

    def _register_array_cmd(self, idx, cmd, log_file=None):
        """
        :param int idx: The index of the job, i.e. which number the job has in the array
        :param str cmd: The command to write
        """
        line = str(idx) + ') ' + cmd
        if log_file:
            line += ' > ' + log_file + ' 2>&1'
        line += '\n' + ';;'
        self.add_line(line)

    def add_line(self, line):
        self.lines.append(line)

    def _start_array(self):
        self.add_line('case $%s in' % self.array_index)

    def _finish_array(self):
        self.add_line('*) echo "Unexpected %s: $%s"' % (self.array_index, self.array_index))
        self.add_line('esac')

    def line_break(self):
        self.lines.append('')

    def save(self):
        """Save self.lines to self.script_name."""
        with open(self.script_name, 'w') as f:
            f.write('\n'.join(self.lines) + '\n')

    def add_header(self):
        """Write a header for a given resource manager. If multiple jobs, split them into a job array."""
        header_mapping = dict(self.cluster_config, log_file=self.log_file, jobs=str(self.array_jobs_written))
        header_lines = list(self.header)

        if self.cluster_config.get('walltime'):
            header_lines.append(self.walltime_header)

        if self.array_jobs_written > 1:
            header_lines.append(self.array_header)

        header_lines.extend(['', 'cd ' + self.working_dir, ''])  # prepend the formatted header
        self.lines = [l.format(**header_mapping) for l in header_lines] + self.lines


class SlurmWriter(ScriptWriter):
    """Writes a Bash script runnable on Slurm"""
    suffix = '.slurm'
    array_index = 'SLURM_ARRAY_TASK_ID'

    header = (
        '#!/bin/bash\n',
        '#SBATCH --job-name="{job_name}"',
        '#SBATCH --cpus-per-task={cpus}',
        '#SBATCH --mem={mem}g',
        '#SBATCH --partition={job_queue}',
        '#SBATCH --output={log_file}'
    )
    walltime_header = '#SBATCH --time={walltime}:00:00'
    array_header = '#SBATCH --array=1-{jobs}'


class PBSWriter(ScriptWriter):
    """Writes a Bash script runnable on PBS"""
    suffix = '.pbs'
    array_index = 'PBS_ARRAY_INDEX'

    header = (
        '#!/bin/bash\n',
        '#PBS -N {job_name}',
        '#PBS -l ncpus={cpus},mem={mem}gb',
        '#PBS -q {job_queue}',
        '#PBS -j oe',
        '#PBS -o {log_file}'
    )
    walltime_header = '#PBS -l walltime={walltime}:00:00'
    array_header = '#PBS -J 1-{jobs}'

    def __init__(self, job_name, working_dir, job_queue, log_commands=True, **cluster_config):
        super().__init__(job_name, working_dir, job_queue, log_commands, **cluster_config)
        if len(self.cluster_config['job_name']) > 15:  # job names longer than 15 chars break PBS
            self.cluster_config['job_name'] = self.cluster_config['job_name'][:15]
